from __future__ import annotations

from contextlib import suppress
from dataclasses import asdict
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, Iterable


class Resolvable:
    """Placeholder for component which can be resolved to the actual object at a later time.
    """
    resolve_field = 'id'

    @classmethod
    def lazy(cls, typ, identifier: str) -> Callable[[], Iterable[Any]]:
        """Factory which returns resolve function for given component.
        """
        def _resolve(items: Iterable[Any]):
            for item in items:
                if isinstance(item, typ) and getattr(item, cls.resolve_field) == identifier:
                    return item

            raise ValueError(f'{typ.__name__} with id "{identifier}" not found')

        return _resolve

    @staticmethod
    def resolve_all(items: Iterable[Any]) -> None:
        """Resolve all given items into real components.
        """
        for item in items:
            with suppress(AttributeError):
                item.resolve(items)

    def resolve(self, items: Iterable[Any]) -> None:
        """Resolve instance to real component.
        """
        for key, value in asdict(self).items():
            with suppress(TypeError):
                setattr(self, key, value(items))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['momos', 'momos.components', 'momos.parser']

package_data = \
{'': ['*'], 'momos': ['libraries/gtest/*', 'templates/gtest/*']}

install_requires = \
['Jinja2>=2.11.2,<3.0.0',
 'case_converter>=1.0.2,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'comment_parser>=1.2.3,<2.0.0',
 'lark>=0.11.1,<0.12.0',
 'lark_parser>=0.11.1,<0.12.0',
 'networkx>=2.5,<3.0',
 'pydot>=1.4.1,<2.0.0',
 'wrapt>=1.12.1,<2.0.0']

entry_points = \
{'console_scripts': ['momos = momos.cli:run']}

setup_kwargs = {
    'name': 'momos',
    'version': '0.1.0',
    'description': '',
    'long_description': None,
    'author': 'codetent',
    'author_email': 'christophsw@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



from __future__ import annotations

import inspect
from functools import partial
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, List


class UnboundFailureMode:
    """Failure mode of a trigger not bound to a trigger instance.
    """
    def __init__(self, generator: Callable, requires: Callable = None, fails: bool = True) -> None:
        self.generator = generator
        self.requires = requires
        self._fails = fails

    @property
    def id(self) -> str:
        return self.generator.__name__

    @property
    def description(self) -> str:
        """Get failure mode description.
        """
        return inspect.getdoc(self.generator).strip()

    def _bind(self, instance: Any) -> FailureMode:
        """Bind failure mode to trigger instance.
        """
        return FailureMode(generator=self.generator, requires=self.requires, fails=self._fails, instance=instance)


class FailureMode(UnboundFailureMode):
    """Instance-bound failure mode.
    """
    def __init__(self, *args, instance: Any = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.instance = instance

    @property
    def arguments(self) -> List[Any]:
        """Get arguments returned by failure mode.
        """
        return self.generator(self.instance) or []

    @property
    def possible(self) -> bool:
        """Check if failure mode is possible for given settings.
        """
        if not callable(self.requires):
            return True
        return self.requires(self.instance)

    @property
    def fails(self) -> bool:
        if not callable(self._fails):
            return self._fails
        return self._fails(self.instance)


class FailureModeResolver:
    """Mixin that resolves specified failure modes.
    """
    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance._failure_modes = {}

        for key, value in cls.__dict__.items():
            if isinstance(value, UnboundFailureMode):
                bound = value._bind(instance)
                setattr(instance, key, bound)
                instance._failure_modes[key] = bound

        return instance


def failure_mode(wrapped: Callable = None, *args, **kwargs) -> UnboundFailureMode:
    """Decorator for creating failure modes inside trigger classes.
    """
    if wrapped is None:
        return partial(failure_mode, *args, **kwargs)

    return UnboundFailureMode(wrapped, *args, **kwargs)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import List

from ..utils import Resolvable
from .trigger import DefaultTrigger, Trigger


@dataclass(frozen=True)
class State:
    id: str
    initial: bool = False


@dataclass
class Transition(Resolvable):
    from_state: State
    to_state: State
    triggers: List[Trigger] = field(default_factory=list)

    @property
    def default_trigger(self):
        try:
            return self.triggers[0]
        except IndexError:
            return DefaultTrigger()
